/*

dynaFastMem

  During the intial compile the BuilderStoreXXX and BuildLoadXXX routines are compiled into the generated code. This is
  done in such a manner that the compiled code only calls the Build routines and then JMPs over an area of reserved 
  memory.  As follows

  call _BuilderStoreWord
  jmp here+0x50
  ..
  ..
  reserved 0x50 bytes
  ..
  ..
  next MIPS instruction

  The builder routine evaluates the Load/Store instruction and compiles custom code into the reserved space, overlaying
  the call to itself.  Thus the next execution of this compiled instruction will directly access the proper translated
  memory area.  The Builder routine accounts for TLB, normal 'virtual' memory and register accesses.  Thus there is a
  slight overhead during the first execution (ie; the time required to compile the custom code), but zero overhead for
  future executions.  Thus sections of the target ROM that are 'run once' initialization type routines will actually
  run somewhat slower, however, these sections are, by nature, rare and do not occur during active gameplay.  While this
  approach is somewhat radical, and self-modifying code is generally a bad idea, we are already creating the code on the 
  fly, so it really makes no difference if we modifiy that code later.  This actually fits the term 'Dynamic Recompilation' 
  better than what is currently the 'industry' standard.

  Assumptions:
	The calculated address for the load/store is based on the designated register and an immediate value.  The assumption
	is that the designated register value will not change by such amount that it would be accessing a different area of
	memory.  This would be possible if the designated base register values were read from a table and the same immediate
	offset were used for all... however, it would still have to 'cross' mapped areas to create a problem.  This is highly 
	unlikely, as it would make little since for the programmer to code in such a fashion.  It is more likely that the 
	designated base register does not change at all, however, it is probably not safe to make this assumption.  Thus the
	translation routine returns a value that added to the base register and offset will equal the proper address.  Thus
	the base register _is_ allowed to change.  This will require and additional add instruction for each use. Thus follows
	the algorithm for the TranslateAddress function:

		TranslatedAddress=PhysicalAddress(VirtualBaseAddress+Immediate)-VirtualAddress; 

	where the routine PhysicalAddress() determines the actual memory mapping and VirtualBaseAddress is the value in the
	designated register.  During the second execution, the address is resolved as follows:  Note that the Immediate value
	has been accounted for in the value of TranslatedAddress.

		ActualAddress=TranslatedAddress+VirtualBaseAddress;

*/
void dynaRuntimeBuilderSB(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer);
void dynaRuntimeBuilderSW(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer);
void dynaRuntimeBuilderSD(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer);
void dynaRuntimeBuilderLD(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer);
void dynaRuntimeBuilderLW(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer);
void dynaRuntimeBuilderLB(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer);
void dynaRuntimeBuilderLBU(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer);
void dynaRuntimeBuilderLH(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer);
void dynaRuntimeBuilderLHU(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer);
void dynaRuntimeBuilderLWC1(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer);
void dynaRuntimeBuilderLDC1(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer);
void dynaRuntimeBuilderLWC2(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer);
void dynaRuntimeBuilderLDC2(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer);
void dynaRuntimeBuilderLL(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer);
void dynaRuntimeBuilderLLD(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer);
void dynaRuntimeBuilderLDL(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer);
void dynaRuntimeBuilderLWL(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer);
void dynaRuntimeBuilderLDR(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer);
void dynaRuntimeBuilderLWR(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer);


extern void helperStoreByte(unsigned char what,unsigned __int32 where);
extern void helperStoreWord(unsigned __int32 what,unsigned __int32 where);
extern void helperStoreDWord(unsigned __int64 what,unsigned __int32 where);
extern unsigned __int64 helperLoadDWord(unsigned __int32 where);
extern unsigned __int32 helperLoadWord(unsigned __int32 where);
extern unsigned __int16 helperLoadHalf(unsigned __int32 where);
extern unsigned char helperLoadByte(unsigned __int32 where);

#define ACCESS_LIMIT 32
#define INIT_ADDRESS 0

#ifdef TRWIN_32
	#define WIDTH 2
#else
	#define WIDTH 1
#endif

unsigned __int16 MAKE_PHYS_ADDR(unsigned char *codeptr,unsigned __int32 Offset,unsigned __int32 Imm)
{
	unsigned __int16 l=0;
	l+=AND_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,0x003fffff);
	l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,Offset+Imm);
	return(l);
}

void dynaTranslateReadAddress(unsigned __int32 VAddr,unsigned __int32 Imm,unsigned __int32 *Addr,unsigned __int32 *mask,bool *IsDirect)
{
	*IsDirect=false;
	if((VAddr&0xa0000000)==0xa0000000)
	{
		*IsDirect=false;
	}
	else
	{
		unsigned __int32 PhysAddr;
		*IsDirect=true;
		if((VAddr&0xffff0000)==0x04000000)
		{
			PhysAddr = (VAddr & 0xffff)+ 0x00390000+(unsigned __int32)dynaRamPtr;
			*Addr=(PhysAddr-(VAddr&0x003fffff));
			*mask=0x003fffff;
			return;
		}
		if((VAddr&0xffff0000)==0x04010000)
		{
			PhysAddr = (VAddr & 0xffff)+ 0x003a0000+(unsigned __int32)dynaRamPtr;
			*Addr=(PhysAddr-(VAddr&0x003fffff));
			*mask=0x003fffff;
			return;
		}
		if((VAddr&0xffff0000)==0x04020000)
		{
			PhysAddr = (VAddr & 0xffff)+ 0x003b0000+(unsigned __int32)dynaRamPtr;
			*Addr=(PhysAddr-(VAddr&0x003fffff));
			*mask=0x003fffff;
			return;
		}
		if((VAddr&0xffff0000)==0x04030000)
		{
			PhysAddr = (VAddr & 0xffff)+ 0x003c0000+(unsigned __int32)dynaRamPtr;
			*Addr=(PhysAddr-(VAddr&0x003fffff));
			*mask=0x003fffff;
			return;
		}
	    unsigned __int32            vaddr = VAddr & 0x1fffffff;
		switch(vaddr & 0x10000000)
		{
			case 0x00000000:
				switch(vaddr & 0x0f000000)
				{
					case 0x00000000:
					case 0x01000000:
					case 0x02000000:
					case 0x03000000:
						{
							if(vaddr < 0x00400000)   /* 4MB */
							{
									PhysAddr=(VAddr&0x003fffff)+(unsigned __int32)dynaRamPtr;
									*Addr=(PhysAddr-(VAddr&0x003fffff));
//									*Addr=(PhysAddr-(VAddr));
									*mask=0x003fffff;
									*IsDirect=true;
									return;
							}
						}
				}
		}
		*IsDirect=false;
	}
}


int dynaCompileBuilderSW(unsigned char *cp,unsigned char op0,unsigned char op1,unsigned __int32 Imm)
{
	unsigned __int16 l=0;
	unsigned char *start=cp;

	l+=INC_PC_COUNTER(cp+l);
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)start);
	l+=PUSH_ALL(cp+l);
	l+=PUSH_STACK_POINTER(cp+l);
	l+=PUSH_DWORD(cp+l,Imm);			//imm
	l+=PUSH_DWORD(cp+l,op1);			//base
	l+=PUSH_DWORD(cp+l,op0);			//what
	l+=PUSH_DWORD(cp+l,(unsigned __int32)start);	//overwrite code area
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)dynaRuntimeBuilderSW);
	l+=CALL_REGISTER(cp+l,NATIVE_0);	// the called address will return back to 'start'+0x50, which will be the
									// next compiled instruction.  This way we don't risk returning to code that
									// we have invalidated with new instructions!
									// note the called function will handle cleaning up the stack
	l+=0x50-l;	// reserve an area of memory
	*(unsigned __int32 *)(cp+0x48)=ACCESS_LIMIT;
	*(unsigned __int32 *)(cp+0x4c)=INIT_ADDRESS;
	return(l);

}

int dynaCompileBuilderSB(unsigned char *cp,unsigned char op0,unsigned char op1,unsigned __int32 Imm)
{
	unsigned __int16 l=0;
	unsigned char *start=cp;

	l+=INC_PC_COUNTER(cp+l);
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)start);
	l+=PUSH_ALL(cp+l);
	l+=PUSH_STACK_POINTER(cp+l);
	l+=PUSH_DWORD(cp+l,Imm);			//imm
	l+=PUSH_DWORD(cp+l,op1);			//base
	l+=PUSH_DWORD(cp+l,op0);			//what
	l+=PUSH_DWORD(cp+l,(unsigned __int32)start);	//overwrite code area
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)dynaRuntimeBuilderSB);
	l+=CALL_REGISTER(cp+l,NATIVE_0);	// the called address will return back to 'start'+0x50, which will be the
									// next compiled instruction.  This way we don't risk returning to code that
									// we have invalidated with new instructions!
									// note the called function will handle cleaning up the stack
	l+=0x50-l;	// reserve an area of memory
	*(unsigned __int32 *)(cp+0x48)=ACCESS_LIMIT;
	*(unsigned __int32 *)(cp+0x4c)=INIT_ADDRESS;
	return(l);

}

int dynaCompileBuilderSD(unsigned char *cp,unsigned char op0,unsigned char op1,unsigned __int32 Imm)
{
	unsigned __int16 l=0;
	unsigned char *start=cp;

	l+=INC_PC_COUNTER(cp+l);
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)start);
	l+=PUSH_ALL(cp+l);
	l+=PUSH_STACK_POINTER(cp+l);
	l+=PUSH_DWORD(cp+l,Imm);			//imm
	l+=PUSH_DWORD(cp+l,op1);			//base
	l+=PUSH_DWORD(cp+l,op0);			//what
	l+=PUSH_DWORD(cp+l,(unsigned __int32)start);	//overwrite code area
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)dynaRuntimeBuilderSD);
	l+=CALL_REGISTER(cp+l,NATIVE_0);	// the called address will return back to 'start'+0x50, which will be the
									// next compiled instruction.  This way we don't risk returning to code that
									// we have invalidated with new instructions!
									// note the called function will handle cleaning up the stack
	l+=0x50-l;	// reserve an area of memory
	*(unsigned __int32 *)(cp+0x48)=ACCESS_LIMIT;
	*(unsigned __int32 *)(cp+0x4c)=INIT_ADDRESS;
	return(l);

}

int dynaCompileBuilderLW(unsigned char *cp,unsigned char op0,unsigned char op1,unsigned __int32 Imm)
{
	unsigned __int16 l=0;
	unsigned char *start=cp;

	l+=INC_PC_COUNTER(cp+l);
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)start);
	l+=PUSH_ALL(cp+l);
	l+=PUSH_STACK_POINTER(cp+l);
	l+=PUSH_DWORD(cp+l,Imm);			//imm
	l+=PUSH_DWORD(cp+l,op1);			//base
	l+=PUSH_DWORD(cp+l,op0);			//what
	l+=PUSH_DWORD(cp+l,(unsigned __int32)start);	//overwrite code area
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)dynaRuntimeBuilderLW);
	l+=CALL_REGISTER(cp+l,NATIVE_0);	// the called address will return back to 'start'+0x50, which will be the
									// next compiled instruction.  This way we don't risk returning to code that
									// we have invalidated with new instructions!
									// note the called function will handle cleaning up the stack
	*(unsigned __int32 *)(cp+0x48)=ACCESS_LIMIT;
	*(unsigned __int32 *)(cp+0x4c)=INIT_ADDRESS;
	l+=0x50-l;	// reserve an area of memory
	return(l);

}


int dynaCompileBuilderLD(unsigned char *cp,unsigned char op0,unsigned char op1,unsigned __int32 Imm)
{
	unsigned __int16 l=0;
	unsigned char *start=cp;

	l+=INC_PC_COUNTER(cp+l);
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)start);
	l+=PUSH_ALL(cp+l);
	l+=PUSH_STACK_POINTER(cp+l);
	l+=PUSH_DWORD(cp+l,Imm);			//imm
	l+=PUSH_DWORD(cp+l,op1);			//base
	l+=PUSH_DWORD(cp+l,op0);			//what
	l+=PUSH_DWORD(cp+l,(unsigned __int32)start);	//overwrite code area
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)dynaRuntimeBuilderLD);
	l+=CALL_REGISTER(cp+l,NATIVE_0);	// the called address will return back to 'start'+0x50, which will be the
									// next compiled instruction.  This way we don't risk returning to code that
									// we have invalidated with new instructions!
									// note the called function will handle cleaning up the stack
	*(unsigned __int32 *)(cp+0x48)=ACCESS_LIMIT;
	*(unsigned __int32 *)(cp+0x4c)=INIT_ADDRESS;
	l+=0x50-l;	// reserve an area of memory
	return(l);

}

int dynaCompileBuilderLDL(unsigned char *cp,unsigned char op0,unsigned char op1,unsigned __int32 Imm)
{
	unsigned __int16 l=0;
	unsigned char *start=cp;

	l+=INC_PC_COUNTER(cp+l);
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)start);
	l+=PUSH_ALL(cp+l);
	l+=PUSH_STACK_POINTER(cp+l);
	l+=PUSH_DWORD(cp+l,Imm);			//imm
	l+=PUSH_DWORD(cp+l,op1);			//base
	l+=PUSH_DWORD(cp+l,op0);			//what
	l+=PUSH_DWORD(cp+l,(unsigned __int32)start);	//overwrite code area
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)dynaRuntimeBuilderLDL);
	l+=CALL_REGISTER(cp+l,NATIVE_0);	// the called address will return back to 'start'+0x50, which will be the
									// next compiled instruction.  This way we don't risk returning to code that
									// we have invalidated with new instructions!
									// note the called function will handle cleaning up the stack
	*(unsigned __int32 *)(cp+0x48)=ACCESS_LIMIT;
	*(unsigned __int32 *)(cp+0x4c)=INIT_ADDRESS;
	l+=0x50-l;	// reserve an area of memory
	return(l);

}


int dynaCompileBuilderLDR(unsigned char *cp,unsigned char op0,unsigned char op1,unsigned __int32 Imm)
{
	unsigned __int16 l=0;
	unsigned char *start=cp;

	l+=INC_PC_COUNTER(cp+l);
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)start);
	l+=PUSH_ALL(cp+l);
	l+=PUSH_STACK_POINTER(cp+l);
	l+=PUSH_DWORD(cp+l,Imm);			//imm
	l+=PUSH_DWORD(cp+l,op1);			//base
	l+=PUSH_DWORD(cp+l,op0);			//what
	l+=PUSH_DWORD(cp+l,(unsigned __int32)start);	//overwrite code area
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)dynaRuntimeBuilderLDR);
	l+=CALL_REGISTER(cp+l,NATIVE_0);	// the called address will return back to 'start'+0x50, which will be the
									// next compiled instruction.  This way we don't risk returning to code that
									// we have invalidated with new instructions!
									// note the called function will handle cleaning up the stack
	*(unsigned __int32 *)(cp+0x48)=ACCESS_LIMIT;
	*(unsigned __int32 *)(cp+0x4c)=INIT_ADDRESS;
	l+=0x50-l;	// reserve an area of memory
	return(l);

}

int dynaCompileBuilderLWL(unsigned char *cp,unsigned char op0,unsigned char op1,unsigned __int32 Imm)
{
	unsigned __int16 l=0;
	unsigned char *start=cp;

	l+=INC_PC_COUNTER(cp+l);
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)start);
	l+=PUSH_ALL(cp+l);
	l+=PUSH_STACK_POINTER(cp+l);
	l+=PUSH_DWORD(cp+l,Imm);			//imm
	l+=PUSH_DWORD(cp+l,op1);			//base
	l+=PUSH_DWORD(cp+l,op0);			//what
	l+=PUSH_DWORD(cp+l,(unsigned __int32)start);	//overwrite code area
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)dynaRuntimeBuilderLWL);
	l+=CALL_REGISTER(cp+l,NATIVE_0);	// the called address will return back to 'start'+0x50, which will be the
									// next compiled instruction.  This way we don't risk returning to code that
									// we have invalidated with new instructions!
									// note the called function will handle cleaning up the stack
	*(unsigned __int32 *)(cp+0x48)=ACCESS_LIMIT;
	*(unsigned __int32 *)(cp+0x4c)=INIT_ADDRESS;
	l+=0x50-l;	// reserve an area of memory
	return(l);

}

int dynaCompileBuilderLWR(unsigned char *cp,unsigned char op0,unsigned char op1,unsigned __int32 Imm)
{
	unsigned __int16 l=0;
	unsigned char *start=cp;

	l+=INC_PC_COUNTER(cp+l);
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)start);
	l+=PUSH_ALL(cp+l);
	l+=PUSH_STACK_POINTER(cp+l);
	l+=PUSH_DWORD(cp+l,Imm);			//imm
	l+=PUSH_DWORD(cp+l,op1);			//base
	l+=PUSH_DWORD(cp+l,op0);			//what
	l+=PUSH_DWORD(cp+l,(unsigned __int32)start);	//overwrite code area
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)dynaRuntimeBuilderLWR);
	l+=CALL_REGISTER(cp+l,NATIVE_0);	// the called address will return back to 'start'+0x50, which will be the
									// next compiled instruction.  This way we don't risk returning to code that
									// we have invalidated with new instructions!
									// note the called function will handle cleaning up the stack
	*(unsigned __int32 *)(cp+0x48)=ACCESS_LIMIT;
	*(unsigned __int32 *)(cp+0x4c)=INIT_ADDRESS;
	l+=0x50-l;	// reserve an area of memory
	return(l);

}

int dynaCompileBuilderLWU(unsigned char *cp,unsigned char op0,unsigned char op1,unsigned __int32 Imm)
{
	unsigned __int16 l=0;
	unsigned char *start=cp;

	l+=INC_PC_COUNTER(cp+l);
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)start);
	l+=PUSH_ALL(cp+l);
	l+=PUSH_STACK_POINTER(cp+l);
	l+=PUSH_DWORD(cp+l,Imm);			//imm
	l+=PUSH_DWORD(cp+l,op1);			//base
	l+=PUSH_DWORD(cp+l,op0);			//what
	l+=PUSH_DWORD(cp+l,(unsigned __int32)start);	//overwrite code area
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)dynaRuntimeBuilderLW);
	l+=CALL_REGISTER(cp+l,NATIVE_0);	// the called address will return back to 'start'+0x50, which will be the
									// next compiled instruction.  This way we don't risk returning to code that
									// we have invalidated with new instructions!
									// note the called function will handle cleaning up the stack
	*(unsigned __int32 *)(cp+0x48)=ACCESS_LIMIT;
	*(unsigned __int32 *)(cp+0x4c)=INIT_ADDRESS;
	l+=0x50-l;	// reserve an area of memory
	return(l);

}

int dynaCompileBuilderLB(unsigned char *cp,unsigned char op0,unsigned char op1,unsigned __int32 Imm)
{
	unsigned __int16 l=0;
	unsigned char *start=cp;

	l+=INC_PC_COUNTER(cp+l);
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)start);
	l+=PUSH_ALL(cp+l);
	l+=PUSH_STACK_POINTER(cp+l);
	l+=PUSH_DWORD(cp+l,Imm);			//imm
	l+=PUSH_DWORD(cp+l,op1);			//base
	l+=PUSH_DWORD(cp+l,op0);			//what
	l+=PUSH_DWORD(cp+l,(unsigned __int32)start);	//overwrite code area
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)dynaRuntimeBuilderLB);
	l+=CALL_REGISTER(cp+l,NATIVE_0);	// the called address will return back to 'start'+0x50, which will be the
									// next compiled instruction.  This way we don't risk returning to code that
									// we have invalidated with new instructions!
									// note the called function will handle cleaning up the stack
	*(unsigned __int32 *)(cp+0x48)=ACCESS_LIMIT;
	*(unsigned __int32 *)(cp+0x4c)=INIT_ADDRESS;
	l+=0x50-l;	// reserve an area of memory
	return(l);

}

int dynaCompileBuilderLH(unsigned char *cp,unsigned char op0,unsigned char op1,unsigned __int32 Imm)
{
	unsigned __int16 l=0;
	unsigned char *start=cp;

	l+=INC_PC_COUNTER(cp+l);
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)start);
	l+=PUSH_ALL(cp+l);
	l+=PUSH_STACK_POINTER(cp+l);
	l+=PUSH_DWORD(cp+l,Imm);			//imm
	l+=PUSH_DWORD(cp+l,op1);			//base
	l+=PUSH_DWORD(cp+l,op0);			//what
	l+=PUSH_DWORD(cp+l,(unsigned __int32)start);	//overwrite code area
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)dynaRuntimeBuilderLH);
	l+=CALL_REGISTER(cp+l,NATIVE_0);	// the called address will return back to 'start'+0x50, which will be the
									// next compiled instruction.  This way we don't risk returning to code that
									// we have invalidated with new instructions!
									// note the called function will handle cleaning up the stack
	*(unsigned __int32 *)(cp+0x48)=ACCESS_LIMIT;
	*(unsigned __int32 *)(cp+0x4c)=INIT_ADDRESS;
	l+=0x50-l;	// reserve an area of memory
	return(l);

}

int dynaCompileBuilderLWC1(unsigned char *cp,unsigned char op0,unsigned char op1,unsigned __int32 Imm)
{
	unsigned __int16 l=0;
	unsigned char *start=cp;

	l+=INC_PC_COUNTER(cp+l);
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)start);
	l+=PUSH_ALL(cp+l);
	l+=PUSH_STACK_POINTER(cp+l);
	l+=PUSH_DWORD(cp+l,Imm);			//imm
	l+=PUSH_DWORD(cp+l,op1);			//base
	l+=PUSH_DWORD(cp+l,op0);			//what
	l+=PUSH_DWORD(cp+l,(unsigned __int32)start);	//overwrite code area
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)dynaRuntimeBuilderLWC1);
	l+=CALL_REGISTER(cp+l,NATIVE_0);	// the called address will return back to 'start'+0x50, which will be the
									// next compiled instruction.  This way we don't risk returning to code that
									// we have invalidated with new instructions!
									// note the called function will handle cleaning up the stack
	*(unsigned __int32 *)(cp+0x48)=ACCESS_LIMIT;
	*(unsigned __int32 *)(cp+0x4c)=INIT_ADDRESS;
	l+=0x50-l;	// reserve an area of memory
	return(l);

}

int dynaCompileBuilderLDC1(unsigned char *cp,unsigned char op0,unsigned char op1,unsigned __int32 Imm)
{
	unsigned __int16 l=0;
	unsigned char *start=cp;

	l+=INC_PC_COUNTER(cp+l);
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)start);
	l+=PUSH_ALL(cp+l);
	l+=PUSH_STACK_POINTER(cp+l);
	l+=PUSH_DWORD(cp+l,Imm);			//imm
	l+=PUSH_DWORD(cp+l,op1);			//base
	l+=PUSH_DWORD(cp+l,op0);			//what
	l+=PUSH_DWORD(cp+l,(unsigned __int32)start);	//overwrite code area
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)dynaRuntimeBuilderLDC1);
	l+=CALL_REGISTER(cp+l,NATIVE_0);	// the called address will return back to 'start'+0x50, which will be the
									// next compiled instruction.  This way we don't risk returning to code that
									// we have invalidated with new instructions!
									// note the called function will handle cleaning up the stack
	*(unsigned __int32 *)(cp+0x48)=ACCESS_LIMIT;
	*(unsigned __int32 *)(cp+0x4c)=INIT_ADDRESS;
	l+=0x50-l;	// reserve an area of memory
	return(l);

}

int dynaCompileBuilderLWC2(unsigned char *cp,unsigned char op0,unsigned char op1,unsigned __int32 Imm)
{
	unsigned __int16 l=0;
	unsigned char *start=cp;

	l+=INC_PC_COUNTER(cp+l);
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)start);
	l+=PUSH_ALL(cp+l);
	l+=PUSH_STACK_POINTER(cp+l);
	l+=PUSH_DWORD(cp+l,Imm);			//imm
	l+=PUSH_DWORD(cp+l,op1);			//base
	l+=PUSH_DWORD(cp+l,op0);			//what
	l+=PUSH_DWORD(cp+l,(unsigned __int32)start);	//overwrite code area
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)dynaRuntimeBuilderLWC2);
	l+=CALL_REGISTER(cp+l,NATIVE_0);	// the called address will return back to 'start'+0x50, which will be the
									// next compiled instruction.  This way we don't risk returning to code that
									// we have invalidated with new instructions!
									// note the called function will handle cleaning up the stack
	*(unsigned __int32 *)(cp+0x48)=ACCESS_LIMIT;
	*(unsigned __int32 *)(cp+0x4c)=INIT_ADDRESS;
	l+=0x50-l;	// reserve an area of memory
	return(l);

}

int dynaCompileBuilderLDC2(unsigned char *cp,unsigned char op0,unsigned char op1,unsigned __int32 Imm)
{
	unsigned __int16 l=0;
	unsigned char *start=cp;

	l+=INC_PC_COUNTER(cp+l);
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)start);
	l+=PUSH_ALL(cp+l);
	l+=PUSH_STACK_POINTER(cp+l);
	l+=PUSH_DWORD(cp+l,Imm);			//imm
	l+=PUSH_DWORD(cp+l,op1);			//base
	l+=PUSH_DWORD(cp+l,op0);			//what
	l+=PUSH_DWORD(cp+l,(unsigned __int32)start);	//overwrite code area
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)dynaRuntimeBuilderLDC2);
	l+=CALL_REGISTER(cp+l,NATIVE_0);	// the called address will return back to 'start'+0x50, which will be the
									// next compiled instruction.  This way we don't risk returning to code that
									// we have invalidated with new instructions!
									// note the called function will handle cleaning up the stack
	*(unsigned __int32 *)(cp+0x48)=ACCESS_LIMIT;
	*(unsigned __int32 *)(cp+0x4c)=INIT_ADDRESS;
	l+=0x50-l;	// reserve an area of memory
	return(l);

}

int dynaCompileBuilderLL(unsigned char *cp,unsigned char op0,unsigned char op1,unsigned __int32 Imm)
{
	unsigned __int16 l=0;
	unsigned char *start=cp;

	l+=INC_PC_COUNTER(cp+l);
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)start);
	l+=PUSH_ALL(cp+l);
	l+=PUSH_STACK_POINTER(cp+l);
	l+=PUSH_DWORD(cp+l,Imm);			//imm
	l+=PUSH_DWORD(cp+l,op1);			//base
	l+=PUSH_DWORD(cp+l,op0);			//what
	l+=PUSH_DWORD(cp+l,(unsigned __int32)start);	//overwrite code area
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)dynaRuntimeBuilderLL);
	l+=CALL_REGISTER(cp+l,NATIVE_0);	// the called address will return back to 'start'+0x50, which will be the
									// next compiled instruction.  This way we don't risk returning to code that
									// we have invalidated with new instructions!
									// note the called function will handle cleaning up the stack
	*(unsigned __int32 *)(cp+0x48)=ACCESS_LIMIT;
	*(unsigned __int32 *)(cp+0x4c)=INIT_ADDRESS;
	l+=0x50-l;	// reserve an area of memory
	return(l);

}


int dynaCompileBuilderLLD(unsigned char *cp,unsigned char op0,unsigned char op1,unsigned __int32 Imm)
{
	unsigned __int16 l=0;
	unsigned char *start=cp;

	l+=INC_PC_COUNTER(cp+l);
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)start);
	l+=PUSH_ALL(cp+l);
	l+=PUSH_STACK_POINTER(cp+l);
	l+=PUSH_DWORD(cp+l,Imm);			//imm
	l+=PUSH_DWORD(cp+l,op1);			//base
	l+=PUSH_DWORD(cp+l,op0);			//what
	l+=PUSH_DWORD(cp+l,(unsigned __int32)start);	//overwrite code area
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)dynaRuntimeBuilderLLD);
	l+=CALL_REGISTER(cp+l,NATIVE_0);	// the called address will return back to 'start'+0x50, which will be the
									// next compiled instruction.  This way we don't risk returning to code that
									// we have invalidated with new instructions!
									// note the called function will handle cleaning up the stack
	*(unsigned __int32 *)(cp+0x48)=ACCESS_LIMIT;
	*(unsigned __int32 *)(cp+0x4c)=INIT_ADDRESS;
	l+=0x50-l;	// reserve an area of memory
	return(l);

}

int dynaCompileBuilderLHU(unsigned char *cp,unsigned char op0,unsigned char op1,unsigned __int32 Imm)
{
	unsigned __int16 l=0;
	unsigned char *start=cp;

	l+=INC_PC_COUNTER(cp+l);
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)start);
	l+=PUSH_ALL(cp+l);
	l+=PUSH_STACK_POINTER(cp+l);
	l+=PUSH_DWORD(cp+l,Imm);			//imm
	l+=PUSH_DWORD(cp+l,op1);			//base
	l+=PUSH_DWORD(cp+l,op0);			//what
	l+=PUSH_DWORD(cp+l,(unsigned __int32)start);	//overwrite code area
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)dynaRuntimeBuilderLHU);
	l+=CALL_REGISTER(cp+l,NATIVE_0);	// the called address will return back to 'start'+0x50, which will be the
									// next compiled instruction.  This way we don't risk returning to code that
									// we have invalidated with new instructions!
									// note the called function will handle cleaning up the stack
	*(unsigned __int32 *)(cp+0x48)=ACCESS_LIMIT;
	*(unsigned __int32 *)(cp+0x4c)=INIT_ADDRESS;
	l+=0x50-l;	// reserve an area of memory
	return(l);

}

int dynaCompileBuilderLBU(unsigned char *cp,unsigned char op0,unsigned char op1,unsigned __int32 Imm)
{
	unsigned __int16 l=0;
	unsigned char *start=cp;

	l+=INC_PC_COUNTER(cp+l);
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)start);
	l+=PUSH_ALL(cp+l);
	l+=PUSH_STACK_POINTER(cp+l);
	l+=PUSH_DWORD(cp+l,Imm);			//imm
	l+=PUSH_DWORD(cp+l,op1);			//base
	l+=PUSH_DWORD(cp+l,op0);			//what
	l+=PUSH_DWORD(cp+l,(unsigned __int32)start);	//overwrite code area
	l+=LOAD_REG_IMM(cp+l,NATIVE_0,(unsigned __int32)dynaRuntimeBuilderLBU);
	l+=CALL_REGISTER(cp+l,NATIVE_0);	// the called address will return back to 'start'+0x50, which will be the
									// next compiled instruction.  This way we don't risk returning to code that
									// we have invalidated with new instructions!
									// note the called function will handle cleaning up the stack
	*(unsigned __int32 *)(cp+0x48)=ACCESS_LIMIT;
	*(unsigned __int32 *)(cp+0x4c)=INIT_ADDRESS;
	l+=0x50-l;	// reserve an area of memory
	return(l);

}

void dynaRuntimeBuilderSW(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer)
{
	unsigned __int16 l=0;
	unsigned __int32 address=m_Reg->r[op1*WIDTH]; //+(int)Imm);
	unsigned __int32 Mask;
	unsigned __int32 value=m_Reg->r[op0*WIDTH];
	unsigned __int32 Offset;
	bool IsDirect;
	helperStoreWord(value,address+Imm);

	bool FlipFlopAddress=false;
	unsigned __int32 LastAddress=*(unsigned __int32  *)((unsigned char *)codeptr+0x4c);
	unsigned __int32 NumAccess=*(unsigned __int32  *)((unsigned char *)codeptr+0x48);
	dynaTranslateReadAddress(address,Imm,&Offset,&Mask,&IsDirect);
	*(unsigned __int32 *)((unsigned char *)codeptr+0x4c)=address;
	NumAccess--;
	*(unsigned __int32 *)((unsigned char *)codeptr+0x48)=NumAccess;
	if(((LastAddress&0xff000000)!=(address&0xff000000))&&(LastAddress!=0))
	{
		IsDirect=false;
		FlipFlopAddress=true;
	}
	if((NumAccess==0)||FlipFlopAddress)
	{
		if(IsDirect)
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=LOAD_REG((unsigned char *)(codeptr+l),op0,NATIVE_0);
			l+=MAKE_PHYS_ADDR((unsigned char *)(codeptr+l),Offset,Imm);
			l+=STORE_MEM_REG((unsigned char *)(codeptr+l),MEM_PTR,NATIVE_0);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
		else
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=LOAD_REG((unsigned char *)(codeptr+l),op0,NATIVE_0);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,Imm);
			l+=PUSH_REGISTER((unsigned char *)(codeptr+l),MEM_PTR);		//where
			l+=PUSH_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);	//what
			l+=LOAD_REG_DWORD((unsigned char *)(codeptr+l),NATIVE_0,(unsigned char *)&helperStoreWordAddress);
			l+=CALL_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=POP_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=POP_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
	}
	_asm
	{
		mov esp,StackPointer	// nasty - get the original stack pointer; ie; before we called the BuilderFunction
								// this 'cleans' the stack.  We put five 32 bit values on the stack and the C++ compiler
								// put some stuff on it as well... we have no way of knowing how much, which is why we 
								// sent the original stackpointer in as a variable.
		popad					// all the original register, cause who knows what C++ did to them.... ;)
		add eax,0x50			// eax contains the starting address... return back to 0x50 past that
		push eax				// new return address
		ret						// hehe... bye!
	}
}

void dynaRuntimeBuilderSB(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer)
{
	unsigned __int16 l=0;
	unsigned __int32 address=m_Reg->r[op1*WIDTH]; //+(int)Imm);
	unsigned __int32 value=m_Reg->r[op0*WIDTH];
	unsigned __int32 Mask;
	unsigned __int32 Offset;
	bool IsDirect;
	helperStoreByte(value,address+Imm);

	bool FlipFlopAddress=false;
	unsigned __int32 LastAddress=*(unsigned __int32  *)((unsigned char *)codeptr+0x4c);
	unsigned __int32 NumAccess=*(unsigned __int32  *)((unsigned char *)codeptr+0x48);
	dynaTranslateReadAddress(address,Imm,&Offset,&Mask,&IsDirect);
	*(unsigned __int32 *)((unsigned char *)codeptr+0x4c)=address;
	NumAccess--;
	*(unsigned __int32 *)((unsigned char *)codeptr+0x48)=NumAccess;
	if(((LastAddress&0xff000000)!=(address&0xff000000))&&(LastAddress!=0))
	{
		IsDirect=false;
		FlipFlopAddress=true;
	}
	if((NumAccess==0)||FlipFlopAddress)
	{
		if(IsDirect)
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=LOAD_REG((unsigned char *)(codeptr+l),op0,NATIVE_0);
			l+=MAKE_PHYS_ADDR((unsigned char *)(codeptr+l),Offset,Imm);
			l+=STORE_MEM_REG((unsigned char *)(codeptr+l),MEM_PTR,NATIVE_0);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
		else
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=LOAD_REG((unsigned char *)(codeptr+l),op0,NATIVE_0);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,Imm);
			l+=PUSH_REGISTER((unsigned char *)(codeptr+l),MEM_PTR);		//where
			l+=PUSH_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);	//what
			l+=LOAD_REG_DWORD((unsigned char *)(codeptr+l),NATIVE_0,(unsigned char *)&helperStoreWordAddress);
			l+=CALL_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=POP_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=POP_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
	}
	_asm
	{
		mov esp,StackPointer	// nasty - get the original stack pointer; ie; before we called the BuilderFunction
								// this 'cleans' the stack.  We put five 32 bit values on the stack and the C++ compiler
								// put some stuff on it as well... we have no way of knowing how much, which is why we 
								// sent the original stackpointer in as a variable.
		popad					// all the original register, cause who knows what C++ did to them.... ;)
		add eax,0x50			// eax contains the starting address... return back to 0x50 past that
		push eax				// new return address
		ret						// hehe... bye!
	}
}

void dynaRuntimeBuilderSD(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer)
{
	unsigned __int16 l=0;
	unsigned __int32 address=m_Reg->r[op1*WIDTH]; //+(int)Imm);
	unsigned __int64 value=*(unsigned __int64 *)&m_Reg->r[op0*WIDTH];
	unsigned __int32 Mask;
	unsigned __int32 Offset;
	bool IsDirect;
	helperStoreDWord(value,address+Imm);

	bool FlipFlopAddress=false;
	unsigned __int32 LastAddress=*(unsigned __int32  *)((unsigned char *)codeptr+0x4c);
	unsigned __int32 NumAccess=*(unsigned __int32  *)((unsigned char *)codeptr+0x48);
	dynaTranslateReadAddress(address,Imm,&Offset,&Mask,&IsDirect);
	*(unsigned __int32 *)((unsigned char *)codeptr+0x4c)=address;
	NumAccess--;
	*(unsigned __int32 *)((unsigned char *)codeptr+0x48)=NumAccess;
	if(((LastAddress&0xff000000)!=(address&0xff000000))&&(LastAddress!=0))
	{
		IsDirect=false;
		FlipFlopAddress=true;
	}
	if((NumAccess==0)||FlipFlopAddress)
	{
		if(IsDirect)
		{

			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=MAKE_PHYS_ADDR((unsigned char *)(codeptr+l),Offset,Imm);
			l+=LOAD_REG_FROM_RBANK((unsigned char *)(codeptr+l),NATIVE_0,op0*8+4);
			l+=STORE_MEM_REG((unsigned char *)(codeptr+l),MEM_PTR,NATIVE_0);
			l+=LOAD_REG_FROM_RBANK((unsigned char *)(codeptr+l),NATIVE_0,op0*8);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,4);
			l+=STORE_MEM_REG((unsigned char *)(codeptr+l),MEM_PTR,NATIVE_0);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
		else
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=LOAD_REG((unsigned char *)(codeptr+l),op0,NATIVE_0);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,Imm);
			l+=PUSH_REGISTER((unsigned char *)(codeptr+l),MEM_PTR);		//where
			l+=LOAD_REG_FROM_RBANK((unsigned char *)(codeptr+l),NATIVE_0,op0*8);
			l+=PUSH_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);	//what
			l+=LOAD_REG_FROM_RBANK((unsigned char *)(codeptr+l),NATIVE_0,op0*8+4);
			l+=PUSH_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);	//what
			l+=LOAD_REG_DWORD((unsigned char *)(codeptr+l),NATIVE_0,(unsigned char *)&helperStoreWordAddress);
			l+=CALL_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=POP_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=POP_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=POP_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
	}
	_asm
	{
		mov esp,StackPointer	// nasty - get the original stack pointer; ie; before we called the BuilderFunction
								// this 'cleans' the stack.  We put five 32 bit values on the stack and the C++ compiler
								// put some stuff on it as well... we have no way of knowing how much, which is why we 
								// sent the original stackpointer in as a variable.
		popad					// all the original register, cause who knows what C++ did to them.... ;)
		add eax,0x50			// eax contains the starting address... return back to 0x50 past that
		push eax				// new return address
		ret						// hehe... bye!
	}
}

void dynaRuntimeBuilderLW(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer)
{
	unsigned __int16 l=0;
	unsigned __int32 address=m_Reg->r[op1*WIDTH]; //+(int)Imm);
	unsigned __int32 value;
	unsigned __int32 Mask;
	unsigned __int32 Offset;
	bool IsDirect;
	value=helperLoadWord(address+Imm);
	m_Reg->r[op0*WIDTH]=value;

	bool FlipFlopAddress=false;
	unsigned __int32 LastAddress=*(unsigned __int32  *)((unsigned char *)codeptr+0x4c);
	unsigned __int32 NumAccess=*(unsigned __int32  *)((unsigned char *)codeptr+0x48);
	dynaTranslateReadAddress(address,Imm,&Offset,&Mask,&IsDirect);
	*(unsigned __int32 *)((unsigned char *)codeptr+0x4c)=address;
	NumAccess--;
	*(unsigned __int32 *)((unsigned char *)codeptr+0x48)=NumAccess;
	if(((LastAddress&0xff000000)!=(address&0xff000000))&&(LastAddress!=0))
	{
		IsDirect=false;
		FlipFlopAddress=true;
	}
	if((NumAccess==0)||FlipFlopAddress)
	{
		if(IsDirect)
		{

			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=MAKE_PHYS_ADDR((unsigned char *)(codeptr+l),Offset,Imm);
			l+=LOAD_MEM_REG((unsigned char *)(codeptr+l),MEM_PTR,NATIVE_0);
			l+=STORE_REG((unsigned char *)(codeptr+l),op0,NATIVE_0);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
		else
		{
			
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,Imm);
			l+=PUSH_REGISTER((unsigned char *)(codeptr+l),MEM_PTR);		//where
			l+=LOAD_REG_DWORD((unsigned char *)(codeptr+l),NATIVE_0,(unsigned char *)&helperLoadWordAddress);
			l+=CALL_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=POP_REGISTER((unsigned char *)(codeptr+l),NATIVE_1);
			l+=STORE_REG((unsigned char *)(codeptr+l),op0,NATIVE_0);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
	}
	_asm
	{
		mov esp,StackPointer	// nasty - get the original stack pointer; ie; before we called the BuilderFunction
								// this 'cleans' the stack.  We put five 32 bit values on the stack and the C++ compiler
								// put some stuff on it as well... we have no way of knowing how much, which is why we 
								// sent the original stackpointer in as a variable.
		popad					// all the original register, cause who knows what C++ did to them.... ;)
		add eax,0x50			// eax contains the starting address... return back to 0x50 past that
		push eax				// new return address
		ret						// hehe... bye!
	}
}


void dynaRuntimeBuilderLD(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer)
{
	unsigned __int16 l=0;
	unsigned __int32 address=m_Reg->r[op1*WIDTH]; //+(int)Imm);
	unsigned __int64 value;
	unsigned __int32 Mask;
	unsigned __int32 Offset;
	bool IsDirect;
	value=helperLoadDWord(address+Imm);
	*(unsigned __int64 *)&m_Reg->r[op0*WIDTH]=value;

	bool FlipFlopAddress=false;
	unsigned __int32 LastAddress=*(unsigned __int32  *)((unsigned char *)codeptr+0x4c);
	unsigned __int32 NumAccess=*(unsigned __int32  *)((unsigned char *)codeptr+0x48);
	dynaTranslateReadAddress(address,Imm,&Offset,&Mask,&IsDirect);
	*(unsigned __int32 *)((unsigned char *)codeptr+0x4c)=address;
	NumAccess--;
	*(unsigned __int32 *)((unsigned char *)codeptr+0x48)=NumAccess;
	if(((LastAddress&0xff000000)!=(address&0xff000000))&&(LastAddress!=0))
	{
		IsDirect=false;
		FlipFlopAddress=true;
	}
	if((NumAccess==0)||FlipFlopAddress)
	{
		if(IsDirect)
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=MAKE_PHYS_ADDR((unsigned char *)(codeptr+l),Offset,Imm);
			l+=LOAD_MEM_REG((unsigned char *)(codeptr+l),MEM_PTR,NATIVE_3);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,4);
			l+=LOAD_MEM_REG((unsigned char *)(codeptr+l),MEM_PTR,NATIVE_0);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_0,op0*8);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_3,op0*8+4);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
		else
		{

			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,Imm);
			l+=PUSH_REGISTER((unsigned char *)(codeptr+l),MEM_PTR);		//where
			l+=LOAD_REG_DWORD((unsigned char *)(codeptr+l),NATIVE_0,(unsigned char *)&helperLoadDWordAddress);
			l+=CALL_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=POP_REGISTER((unsigned char *)(codeptr+l),NATIVE_1);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_0,op0*8);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_3,op0*8+4);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
	}
	_asm
	{
		mov esp,StackPointer	// nasty - get the original stack pointer; ie; before we called the BuilderFunction
								// this 'cleans' the stack.  We put five 32 bit values on the stack and the C++ compiler
								// put some stuff on it as well... we have no way of knowing how much, which is why we 
								// sent the original stackpointer in as a variable.
		popad					// all the original register, cause who knows what C++ did to them.... ;)
		add eax,0x50			// eax contains the starting address... return back to 0x50 past that
		push eax				// new return address
		ret						// hehe... bye!
	}
}


void dynaRuntimeBuilderLB(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer)
{
	unsigned __int16 l=0;
	bool FlipFlopAddress=false;	
	unsigned __int32 address=m_Reg->r[op1*WIDTH]+Imm;
	unsigned __int32 LastAddress=*(unsigned __int32  *)((unsigned char *)codeptr+0x4c);
	unsigned __int32 NumAccess=*(unsigned __int32  *)((unsigned char *)codeptr+0x48);
	__int32 value;
	unsigned __int32 Mask;
	unsigned __int32 Offset;
	bool IsDirect;
	value=(__int32)(char)helperLoadByte(address);
	m_Reg->r[op0*WIDTH]=value;
	*(unsigned __int32 *)((unsigned char *)codeptr+0x4c)=address;
	NumAccess--;
	*(unsigned __int32 *)((unsigned char *)codeptr+0x48)=NumAccess;
	dynaTranslateReadAddress(address,Imm,&Offset,&Mask,&IsDirect);
	if(((LastAddress&0xff000000)!=(address&0xff000000))&&(LastAddress!=0))
	{
		IsDirect=false;
		FlipFlopAddress=true;
	}
	if((NumAccess==0)||FlipFlopAddress)
	{
		if(IsDirect)
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=MAKE_PHYS_ADDR((unsigned char *)(codeptr+l),Offset,Imm);
			l+=XOR_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,3);
			l+=LOAD_MEM_REG((unsigned char *)(codeptr+l),MEM_PTR,NATIVE_0);
			l+=CONVERT_TO_WORD((unsigned char *)(codeptr+l));
			l+=CONVERT_TO_DWORD((unsigned char *)(codeptr+l));
			l+=STORE_REG((unsigned char *)(codeptr+l),op0,NATIVE_0);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
		else
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,Imm);
			l+=PUSH_REGISTER((unsigned char *)(codeptr+l),MEM_PTR);		//where
			l+=LOAD_REG_DWORD((unsigned char *)(codeptr+l),NATIVE_0,(unsigned char *)&helperLoadByteAddress);
			l+=CALL_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=POP_REGISTER((unsigned char *)(codeptr+l),NATIVE_1);
			l+=CONVERT_TO_WORD((unsigned char *)(codeptr+l));
			l+=CONVERT_TO_DWORD((unsigned char *)(codeptr+l));
			l+=STORE_REG((unsigned char *)(codeptr+l),op0,NATIVE_0);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
	}
	_asm
	{
		mov esp,StackPointer	// nasty - get the original stack pointer; ie; before we called the BuilderFunction
								// this 'cleans' the stack.  We put five 32 bit values on the stack and the C++ compiler
								// put some stuff on it as well... we have no way of knowing how much, which is why we 
								// sent the original stackpointer in as a variable.
		popad					// all the original register, cause who knows what C++ did to them.... ;)
		add eax,0x50			// eax contains the starting address... return back to 0x50 past that
		push eax				// new return address
		ret						// hehe... bye!
	}
}


void dynaRuntimeBuilderLBU(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer)
{
	unsigned __int16 l=0;
	unsigned __int32 address=m_Reg->r[op1*WIDTH]; //+(int)Imm);
	unsigned __int32 Mask;
	unsigned __int32 value;
	unsigned __int32 Offset;
	bool IsDirect;
	value=(unsigned __int32)(unsigned char)helperLoadByte(address+Imm);
	m_Reg->r[op0*WIDTH]=value;

	bool FlipFlopAddress=false;
	unsigned __int32 LastAddress=*(unsigned __int32  *)((unsigned char *)codeptr+0x4c);
	unsigned __int32 NumAccess=*(unsigned __int32  *)((unsigned char *)codeptr+0x48);
	dynaTranslateReadAddress(address,Imm,&Offset,&Mask,&IsDirect);
	*(unsigned __int32 *)((unsigned char *)codeptr+0x4c)=address;
	NumAccess--;
	*(unsigned __int32 *)((unsigned char *)codeptr+0x48)=NumAccess;
	if(((LastAddress&0xff000000)!=(address&0xff000000))&&(LastAddress!=0))
	{
		IsDirect=false;
		FlipFlopAddress=true;
	}
	if((NumAccess==0)||FlipFlopAddress)
	{
		if(IsDirect)
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=MAKE_PHYS_ADDR((unsigned char *)(codeptr+l),Offset,Imm);
			l+=XOR_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,3);
			l+=LOAD_MEM_REG((unsigned char *)(codeptr+l),MEM_PTR,NATIVE_0);
			l+=AND_REG_IMM((unsigned char *)(codeptr+l),NATIVE_0,0x000000ff);
			l+=STORE_REG((unsigned char *)(codeptr+l),op0,NATIVE_0);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
		else
		{

			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,Imm);
			l+=PUSH_REGISTER((unsigned char *)(codeptr+l),MEM_PTR);		//where
			l+=LOAD_REG_DWORD((unsigned char *)(codeptr+l),NATIVE_0,(unsigned char *)&helperLoadByteAddress);
			l+=CALL_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=POP_REGISTER((unsigned char *)(codeptr+l),NATIVE_1);
			l+=AND_REG_IMM((unsigned char *)(codeptr+l),NATIVE_0,0x000000ff);
			l+=STORE_REG((unsigned char *)(codeptr+l),op0,NATIVE_0);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
	}
	_asm
	{
		mov esp,StackPointer	// nasty - get the original stack pointer; ie; before we called the BuilderFunction
								// this 'cleans' the stack.  We put five 32 bit values on the stack and the C++ compiler
								// put some stuff on it as well... we have no way of knowing how much, which is why we 
								// sent the original stackpointer in as a variable.
		popad					// all the original register, cause who knows what C++ did to them.... ;)
		add eax,0x50			// eax contains the starting address... return back to 0x50 past that
		push eax				// new return address
		ret						// hehe... bye!
	}
}


void dynaRuntimeBuilderLH(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer)
{
	unsigned __int16 l=0;
	bool FlipFlopAddress=false;	
	unsigned __int32 address=m_Reg->r[op1*WIDTH]+Imm;
	unsigned __int32 LastAddress=*(unsigned __int32  *)((unsigned char *)codeptr+0x4c);
	unsigned __int32 NumAccess=*(unsigned __int32  *)((unsigned char *)codeptr+0x48);
	__int32 value;
	unsigned __int32 Mask;
	unsigned __int32 Offset;
	bool IsDirect;
	value=(__int32)(short)helperLoadHalf(address);
	m_Reg->r[op0*WIDTH]=value;
	*(unsigned __int32 *)((unsigned char *)codeptr+0x4c)=address;
	NumAccess--;
	*(unsigned __int32 *)((unsigned char *)codeptr+0x48)=NumAccess;
	dynaTranslateReadAddress(address,Imm,&Offset,&Mask,&IsDirect);
	if(((LastAddress&0xff000000)!=(address&0xff000000))&&(LastAddress!=0))
	{
		IsDirect=false;
		FlipFlopAddress=true;
	}
	if((NumAccess==0)||FlipFlopAddress)
	{
		if(IsDirect)
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=MAKE_PHYS_ADDR((unsigned char *)(codeptr+l),Offset,Imm);
			l+=XOR_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,2);
			l+=LOAD_MEM_REG((unsigned char *)(codeptr+l),MEM_PTR,NATIVE_0);
			l+=CONVERT_TO_DWORD((unsigned char *)(codeptr+l));
			l+=STORE_REG((unsigned char *)(codeptr+l),op0,NATIVE_0);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
		else
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,Imm);
			l+=PUSH_REGISTER((unsigned char *)(codeptr+l),MEM_PTR);		
			l+=LOAD_REG_DWORD((unsigned char *)(codeptr+l),NATIVE_0,(unsigned char *)&helperLoadHalfAddress);
			l+=CALL_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=POP_REGISTER((unsigned char *)(codeptr+l),NATIVE_1);
			l+=CONVERT_TO_DWORD((unsigned char *)(codeptr+l));
			l+=STORE_REG((unsigned char *)(codeptr+l),op0,NATIVE_0);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
	}
	_asm
	{
		mov esp,StackPointer	// nasty - get the original stack pointer; ie; before we called the BuilderFunction
								// this 'cleans' the stack.  We put five 32 bit values on the stack and the C++ compiler
								// put some stuff on it as well... we have no way of knowing how much, which is why we 
								// sent the original stackpointer in as a variable.
		popad					// all the original register, cause who knows what C++ did to them.... ;)
		add eax,0x50			// eax contains the starting address... return back to 0x50 past that
		push eax				// new return address
		ret						// hehe... bye!
	}
}


void dynaRuntimeBuilderLHU(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer)
{
	unsigned __int16 l=0;
	bool FlipFlopAddress=false;	
	unsigned __int32 address=m_Reg->r[op1*WIDTH]+Imm;
	unsigned __int32 LastAddress=*(unsigned __int32  *)((unsigned char *)codeptr+0x4c);
	unsigned __int32 NumAccess=*(unsigned __int32  *)((unsigned char *)codeptr+0x48);
	__int32 value;
	unsigned __int32 Mask;
	unsigned __int32 Offset;
	bool IsDirect;
	value=(unsigned __int32)(unsigned __int16)helperLoadHalf(address);
	m_Reg->r[op0*WIDTH]=value;
	*(unsigned __int32 *)((unsigned char *)codeptr+0x4c)=address;
	NumAccess--;
	*(unsigned __int32 *)((unsigned char *)codeptr+0x48)=NumAccess;
	dynaTranslateReadAddress(address,Imm,&Offset,&Mask,&IsDirect);
	if(((LastAddress&0xff000000)!=(address&0xff000000))&&(LastAddress!=0))
	{
		IsDirect=false;
		FlipFlopAddress=true;
	}
	if((NumAccess==0)||FlipFlopAddress)
	{
		if(IsDirect)//&&(Offset!=(unsigned __int32)dynaRamPtr))
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=MAKE_PHYS_ADDR((unsigned char *)(codeptr+l),Offset,Imm);
			l+=XOR_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,2);
			l+=LOAD_MEM_REG((unsigned char *)(codeptr+l),MEM_PTR,NATIVE_0);
			l+=AND_REG_IMM((unsigned char *)(codeptr+l),NATIVE_0,0x0000ffff);
			l+=STORE_REG((unsigned char *)(codeptr+l),op0,NATIVE_0);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
		else
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,Imm);
			l+=PUSH_REGISTER((unsigned char *)(codeptr+l),MEM_PTR);		//where
			l+=LOAD_REG_DWORD((unsigned char *)(codeptr+l),NATIVE_0,(unsigned char *)&helperLoadHalfAddress);
			l+=CALL_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=POP_REGISTER((unsigned char *)(codeptr+l),NATIVE_1);
			l+=AND_REG_IMM((unsigned char *)(codeptr+l),NATIVE_0,0x0000ffff);
//			l+=CONVERT_TO_DWORD((unsigned char *)(codeptr+l));
			l+=STORE_REG((unsigned char *)(codeptr+l),op0,NATIVE_0);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
	}
	_asm
	{
		mov esp,StackPointer	// nasty - get the original stack pointer; ie; before we called the BuilderFunction
								// this 'cleans' the stack.  We put five 32 bit values on the stack and the C++ compiler
								// put some stuff on it as well... we have no way of knowing how much, which is why we 
								// sent the original stackpointer in as a variable.
		popad					// all the original register, cause who knows what C++ did to them.... ;)
		add eax,0x50			// eax contains the starting address... return back to 0x50 past that
		push eax				// new return address
		ret						// hehe... bye!
	}
}


void dynaRuntimeBuilderLWC1(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer)
{
	unsigned __int16 l=0;
	bool FlipFlopAddress=false;	
	unsigned __int32 address=m_Reg->r[op1*WIDTH]+Imm;
	unsigned __int32 LastAddress=*(unsigned __int32  *)((unsigned char *)codeptr+0x4c);
	unsigned __int32 NumAccess=*(unsigned __int32  *)((unsigned char *)codeptr+0x48);
	__int32 value;
	unsigned __int32 Mask;
	unsigned __int32 Offset;
	bool IsDirect;
	value=(__int32)helperLoadWord(address);
	m_Reg->FPR[op0]=value;
	*(unsigned __int32 *)((unsigned char *)codeptr+0x4c)=address;
	NumAccess--;
	*(unsigned __int32 *)((unsigned char *)codeptr+0x48)=NumAccess;
	dynaTranslateReadAddress(address,Imm,&Offset,&Mask,&IsDirect);
	if(((LastAddress&0xff000000)!=(address&0xff000000))&&(LastAddress!=0))
	{
		IsDirect=false;
		FlipFlopAddress=true;
	}
	if((NumAccess==0)||FlipFlopAddress)
	{
		if(IsDirect)//&&(Offset!=(unsigned __int32)dynaRamPtr))
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=MAKE_PHYS_ADDR((unsigned char *)(codeptr+l),Offset,Imm);
			l+=LOAD_MEM_REG((unsigned char *)(codeptr+l),MEM_PTR,NATIVE_0);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_0,_FPR+op0*4);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
		else
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,Imm);
			l+=PUSH_REGISTER((unsigned char *)(codeptr+l),MEM_PTR);		//where
			l+=LOAD_REG_DWORD((unsigned char *)(codeptr+l),NATIVE_0,(unsigned char *)&helperLoadWordAddress);
			l+=CALL_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=POP_REGISTER((unsigned char *)(codeptr+l),NATIVE_1);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_0,_FPR+op0*4);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
	}
	_asm
	{
		mov esp,StackPointer	// nasty - get the original stack pointer; ie; before we called the BuilderFunction
								// this 'cleans' the stack.  We put five 32 bit values on the stack and the C++ compiler
								// put some stuff on it as well... we have no way of knowing how much, which is why we 
								// sent the original stackpointer in as a variable.
		popad					// all the original register, cause who knows what C++ did to them.... ;)
		add eax,0x50			// eax contains the starting address... return back to 0x50 past that
		push eax				// new return address
		ret						// hehe... bye!
	}
}


void dynaRuntimeBuilderLL(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer)
{
	unsigned __int16 l=0;
	bool FlipFlopAddress=false;	
	unsigned __int32 address=m_Reg->r[op1*WIDTH]+Imm;
	unsigned __int32 LastAddress=*(unsigned __int32  *)((unsigned char *)codeptr+0x4c);
	unsigned __int32 NumAccess=*(unsigned __int32  *)((unsigned char *)codeptr+0x48);
	__int32 value;
	unsigned __int32 Mask;
	unsigned __int32 Offset;
	bool IsDirect;
	value=(__int32)helperLoadWord(address);

	m_Reg->r[op0*WIDTH]=value;
	m_Reg->CPR[0][LLADDR] = (__int64)address;
	m_Reg->Llbit = 1;

	*(unsigned __int32 *)((unsigned char *)codeptr+0x4c)=address;
	NumAccess--;
	*(unsigned __int32 *)((unsigned char *)codeptr+0x48)=NumAccess;
	dynaTranslateReadAddress(address,Imm,&Offset,&Mask,&IsDirect);
	if(((LastAddress&0xff000000)!=(address&0xff000000))&&(LastAddress!=0))
	{
		IsDirect=false;
		FlipFlopAddress=true;
	}
	if((NumAccess==0)||FlipFlopAddress)
	{
		if(IsDirect)//&&(Offset!=(unsigned __int32)dynaRamPtr))
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,Imm);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),MEM_PTR,CPR0+17*8);
			l+=LOAD_REG_IMM((unsigned char *)(codeptr+l),NATIVE_0,1);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_0,1972);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,Offset);
			l+=LOAD_MEM_REG((unsigned char *)(codeptr+l),MEM_PTR,NATIVE_0);
			l+=STORE_REG((unsigned char *)(codeptr+l),op0,NATIVE_0);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
		else
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,Imm);
			l+=PUSH_REGISTER((unsigned char *)(codeptr+l),MEM_PTR);		//where
			l+=LOAD_REG_DWORD((unsigned char *)(codeptr+l),NATIVE_0,(unsigned char *)&helperLoadWordAddress);
			l+=CALL_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=POP_REGISTER((unsigned char *)(codeptr+l),NATIVE_1);
			l+=STORE_REG((unsigned char *)(codeptr+l),op0,NATIVE_0);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
	}
	_asm
	{
		mov esp,StackPointer	// nasty - get the original stack pointer; ie; before we called the BuilderFunction
								// this 'cleans' the stack.  We put five 32 bit values on the stack and the C++ compiler
								// put some stuff on it as well... we have no way of knowing how much, which is why we 
								// sent the original stackpointer in as a variable.
		popad					// all the original register, cause who knows what C++ did to them.... ;)
		add eax,0x50			// eax contains the starting address... return back to 0x50 past that
		push eax				// new return address
		ret						// hehe... bye!
	}
}


void dynaRuntimeBuilderLLD(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer)
{
	unsigned __int16 l=0;
	bool FlipFlopAddress=false;	
	unsigned __int32 address=m_Reg->r[op1*WIDTH]+Imm;
	unsigned __int32 LastAddress=*(unsigned __int32  *)((unsigned char *)codeptr+0x4c);
	unsigned __int32 NumAccess=*(unsigned __int32  *)((unsigned char *)codeptr+0x48);
	__int64 value;
	unsigned __int32 Mask;
	unsigned __int32 Offset;
	bool IsDirect;
	value=(__int32)helperLoadDWord(address);

	*(__int64 *)&m_Reg->r[op0*WIDTH]=value;
	m_Reg->CPR[0][LLADDR] = (__int64)address;
	m_Reg->Llbit = 1;

	*(unsigned __int32 *)((unsigned char *)codeptr+0x4c)=address;
	NumAccess--;
	*(unsigned __int32 *)((unsigned char *)codeptr+0x48)=NumAccess;
	dynaTranslateReadAddress(address,Imm,&Offset,&Mask,&IsDirect);
	if(((LastAddress&0xff000000)!=(address&0xff000000))&&(LastAddress!=0))
	{
		IsDirect=false;
		FlipFlopAddress=true;
	}
	if((NumAccess==0)||FlipFlopAddress)
	{
		if(IsDirect)//&&(Offset!=(unsigned __int32)dynaRamPtr))
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,Imm);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),MEM_PTR,CPR0+17*8);
			l+=LOAD_REG_IMM((unsigned char *)(codeptr+l),NATIVE_0,1);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_0,1972);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,Offset);
			l+=LOAD_MEM_REG((unsigned char *)(codeptr+l),MEM_PTR,NATIVE_3);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,4);
			l+=LOAD_MEM_REG((unsigned char *)(codeptr+l),MEM_PTR,NATIVE_0);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_0,op0*8);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_3,op0*8+4);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
		else
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,Imm);
			l+=PUSH_REGISTER((unsigned char *)(codeptr+l),MEM_PTR);		//where
			l+=LOAD_REG_DWORD((unsigned char *)(codeptr+l),NATIVE_0,(unsigned char *)&helperLoadDWordAddress);
			l+=CALL_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=POP_REGISTER((unsigned char *)(codeptr+l),NATIVE_1);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_0,_FPR+op0*4);	//this is wrong... but never called
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
	}
	_asm
	{
		mov esp,StackPointer	// nasty - get the original stack pointer; ie; before we called the BuilderFunction
								// this 'cleans' the stack.  We put five 32 bit values on the stack and the C++ compiler
								// put some stuff on it as well... we have no way of knowing how much, which is why we 
								// sent the original stackpointer in as a variable.
		popad					// all the original register, cause who knows what C++ did to them.... ;)
		add eax,0x50			// eax contains the starting address... return back to 0x50 past that
		push eax				// new return address
		ret						// hehe... bye!
	}
}


void dynaRuntimeBuilderLDC1(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer)
{
	unsigned __int16 l=0;
	bool FlipFlopAddress=false;	
	unsigned __int32 address=m_Reg->r[op1*WIDTH]+Imm;
	unsigned __int32 LastAddress=*(unsigned __int32  *)((unsigned char *)codeptr+0x4c);
	unsigned __int32 NumAccess=*(unsigned __int32  *)((unsigned char *)codeptr+0x48);
	__int64 value;
	unsigned __int32 Mask;
	unsigned __int32 Offset;
	bool IsDirect;
	value=(__int64)helperLoadDWord(address);
	if(m_Reg->CPR[0][STATUS] & 0x04000000)
	{
		m_Reg->FPR[op0]=(__int32)value;
	}
	else
	{
		m_Reg->FPR[op0]=(__int32)value;
		m_Reg->FPR[op0+1]=(__int32)(value>>32);
	}

	*(unsigned __int32 *)((unsigned char *)codeptr+0x4c)=address;
	NumAccess--;
	*(unsigned __int32 *)((unsigned char *)codeptr+0x48)=NumAccess;
	dynaTranslateReadAddress(address,Imm,&Offset,&Mask,&IsDirect);
	if(((LastAddress&0xff000000)!=(address&0xff000000))&&(LastAddress!=0))
	{
		IsDirect=false;
		FlipFlopAddress=true;
	}
	if((NumAccess==0)||FlipFlopAddress)
	{
		if(IsDirect)//&&(Offset!=(unsigned __int32)dynaRamPtr))
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=MAKE_PHYS_ADDR((unsigned char *)(codeptr+l),Offset,Imm);
			l+=LOAD_MEM_REG((unsigned char *)(codeptr+l),MEM_PTR,NATIVE_0);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_0,_FPR+op0*4+4);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,4);
			l+=LOAD_MEM_REG((unsigned char *)(codeptr+l),MEM_PTR,NATIVE_0);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_0,_FPR+op0*4);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
		else
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,Imm);
			l+=PUSH_REGISTER((unsigned char *)(codeptr+l),MEM_PTR);		//where
			l+=LOAD_REG_DWORD((unsigned char *)(codeptr+l),NATIVE_0,(unsigned char *)&helperLoadWordAddress);
			l+=CALL_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=POP_REGISTER((unsigned char *)(codeptr+l),NATIVE_1);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_0,_FPR+op0*4);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
	}
	_asm
	{
		mov esp,StackPointer	// nasty - get the original stack pointer; ie; before we called the BuilderFunction
								// this 'cleans' the stack.  We put five 32 bit values on the stack and the C++ compiler
								// put some stuff on it as well... we have no way of knowing how much, which is why we 
								// sent the original stackpointer in as a variable.
		popad					// all the original register, cause who knows what C++ did to them.... ;)
		add eax,0x50			// eax contains the starting address... return back to 0x50 past that
		push eax				// new return address
		ret						// hehe... bye!
	}
}

void dynaHelperLDL(unsigned __int32 address,unsigned __int32 op0)
{
	unsigned __int32    offset;
	unsigned __int64   data;
	
	offset = (unsigned __int32)address;
	
	data=helperLoadDWord(offset & 0xfffffff8);

	switch(offset % 8)
	{
	case 0:
		*(unsigned __int64 *)&m_Reg->r[op0*WIDTH] = data;
		break;
		
	case 1:
		*(unsigned __int64 *)&m_Reg->r[op0*WIDTH]  = (*(unsigned __int64 *)&m_Reg->r[op0*WIDTH] & 0x00000000000000ff) | (data << 8);
		break;
		
	case 2:
		*(unsigned __int64 *)&m_Reg->r[op0*WIDTH]  = (*(unsigned __int64 *)&m_Reg->r[op0*WIDTH] & 0x000000000000ffff) | (data << 16);
		break;
		
	case 3:
		*(unsigned __int64 *)&m_Reg->r[op0*WIDTH]  = (*(unsigned __int64 *)&m_Reg->r[op0*WIDTH] & 0x0000000000ffffff) | (data << 24);
		break;
		
	case 4:
		*(unsigned __int64 *)&m_Reg->r[op0*WIDTH]  = (*(unsigned __int64 *)&m_Reg->r[op0*WIDTH] & 0x00000000ffffffff) | (data << 32);
		break;
		
	case 5:
		*(unsigned __int64 *)&m_Reg->r[op0*WIDTH]  = (*(unsigned __int64 *)&m_Reg->r[op0*WIDTH] & 0x000000ffffffffff) | (data << 40);
		break;
		
	case 6:
		*(unsigned __int64 *)&m_Reg->r[op0*WIDTH]  = (*(unsigned __int64 *)&m_Reg->r[op0*WIDTH] & 0x0000ffffffffffff) | (data << 48);
		break;
		
	case 7:
		*(unsigned __int64 *)&m_Reg->r[op0*WIDTH]  = (*(unsigned __int64 *)&m_Reg->r[op0*WIDTH] & 0x00ffffffffffffff) | (data << 56);
		break;
		
	} 
}

void dynaHelperLWL(unsigned __int32 address,unsigned __int32 op0)
{
	unsigned __int32    offset;
	unsigned __int32   data;
	
	offset = (unsigned __int32)address;
	
	data=helperLoadWord(offset & 0xfffffffc);

	switch(offset % 4)
	{
	case 0:
		*(__int64 *)&m_Reg->r[op0*WIDTH] = (__int64)(__int32)data;
		break;
		
	case 1:
		*(__int64 *)&m_Reg->r[op0*WIDTH]  = (__int64)(__int32)((m_Reg->r[op0*WIDTH] & 0x000000ff) | (data << 8));
		break;
		
	case 2:
		*(__int64 *)&m_Reg->r[op0*WIDTH]  = (__int64)(__int32)((m_Reg->r[op0*WIDTH] & 0x0000ffff) | (data << 16));
		break;
		
	case 3:
		*(__int64 *)&m_Reg->r[op0*WIDTH]  = (__int64)(__int32)((m_Reg->r[op0*WIDTH] & 0x00ffffff) | (data << 24));
		break;
	}
}

void dynaHelperLDR(unsigned __int32 address,unsigned __int32 op0)
{
	unsigned __int32    offset;
	unsigned __int64   data;
	
	offset = (unsigned __int32)address;
	
	data=helperLoadDWord(offset & 0xfffffff8);

	switch(offset % 8)
	{
	case 7:
		*(unsigned __int64 *)&m_Reg->r[op0*WIDTH] = data;
		break;
		
	case 6:
		*(unsigned __int64 *)&m_Reg->r[op0*WIDTH]  = (*(unsigned __int64 *)&m_Reg->r[op0*WIDTH] & 0xff00000000000000) | (data >> 8);
		break;
		
	case 5:
		*(unsigned __int64 *)&m_Reg->r[op0*WIDTH]  = (*(unsigned __int64 *)&m_Reg->r[op0*WIDTH] & 0xffff000000000000) | (data >> 16);
		break;
		
	case 4:
		*(unsigned __int64 *)&m_Reg->r[op0*WIDTH]  = (*(unsigned __int64 *)&m_Reg->r[op0*WIDTH] & 0xffffff0000000000) | (data >> 24);
		break;
		
	case 3:
		*(unsigned __int64 *)&m_Reg->r[op0*WIDTH]  = (*(unsigned __int64 *)&m_Reg->r[op0*WIDTH] & 0xffffffff00000000) | (data >> 32);
		break;
		
	case 2:
		*(unsigned __int64 *)&m_Reg->r[op0*WIDTH]  = (*(unsigned __int64 *)&m_Reg->r[op0*WIDTH] & 0xffffffffff000000) | (data >> 40);
		break;
		
	case 1:
		*(unsigned __int64 *)&m_Reg->r[op0*WIDTH]  = (*(unsigned __int64 *)&m_Reg->r[op0*WIDTH] & 0xffffffffffff0000) | (data >> 48);
		break;
		
	case 0:
		*(unsigned __int64 *)&m_Reg->r[op0*WIDTH]  = (*(unsigned __int64 *)&m_Reg->r[op0*WIDTH] & 0xffffffffffffff00) | (data >> 56);
		break;
		
	}
}

void dynaHelperLWR(unsigned __int32 address,unsigned __int32 op0)
{
	unsigned __int32    offset;
	unsigned __int32   data;
	
	offset = (unsigned __int32)address;
	
	data=helperLoadWord(offset & 0xfffffffc);

	switch(offset % 4)
	{
	case 0:
		*(__int64 *)&m_Reg->r[op0*WIDTH]  = (*(__int64 *)&m_Reg->r[op0*WIDTH] & 0xffffff00) | (data >> 24);
		break;
		
	case 1:
		*(__int64 *)&m_Reg->r[op0*WIDTH]  = (*(__int64 *)&m_Reg->r[op0*WIDTH] & 0xffff0000) | (data >> 16);
		break;
		
	case 2:
		*(__int64 *)&m_Reg->r[op0*WIDTH]  = (*(__int64 *)&m_Reg->r[op0*WIDTH] & 0xff000000) | (data >> 8);
		break;
		
	case 3:
		*(__int64 *)&m_Reg->r[op0*WIDTH]  = (__int64)(__int32)data;
		break;
	}
}

void dynaRuntimeBuilderLDL(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer)
{
	unsigned __int16 l=0;
	bool FlipFlopAddress=false;	
	unsigned __int32 address=m_Reg->r[op1*WIDTH]+Imm;
	unsigned __int32 LastAddress=*(unsigned __int32  *)((unsigned char *)codeptr+0x4c);
	unsigned __int32 NumAccess=*(unsigned __int32  *)((unsigned char *)codeptr+0x48);
	__int32 value;
	unsigned __int32 Mask;
	unsigned __int32 Offset;
	bool IsDirect;

	dynaHelperLDL(address,op0);

	*(unsigned __int32 *)((unsigned char *)codeptr+0x4c)=address;
	NumAccess--;
	*(unsigned __int32 *)((unsigned char *)codeptr+0x48)=NumAccess;
	dynaTranslateReadAddress(address,Imm,&Offset,&Mask,&IsDirect);
	if(((LastAddress&0xff000000)!=(address&0xff000000))&&(LastAddress!=0))
	{
		IsDirect=false;
		FlipFlopAddress=true;
	}
	if((NumAccess==0)||FlipFlopAddress)
	{
		if(IsDirect)//&&(Offset!=(unsigned __int32)dynaRamPtr))
		{
			int Part1,Part2,Count;
			unsigned __int32 Offset2;
			Offset2=Imm&7;
			Count=8-(int)Offset2;
			Part1=Count-4;
			if(Count>=4) Part2=4;
			else Part2=Count;
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			if(Part1>0)
			{
				unsigned __int16 l=0;
				l+=PUSH_ALL((unsigned char *)(codeptr+l));
				l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
				l+=CLD((unsigned char *)(codeptr+l));
				l+=MAKE_PHYS_ADDR((unsigned char *)(codeptr+l),Offset,Imm);
				unsigned __int32 tmp=op0*8;
				l+=NATIVE_REG_TO_REG((unsigned char *)(codeptr+l),PC_PTR,REG_PTR);
				l+=ADD_REG_IMM((unsigned char *)(codeptr+l),PC_PTR,tmp);
				l+=PUSH_REGISTER((unsigned char *)(codeptr+l),MEM_PTR);
				tmp=4;
				l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,tmp);		// src+4 (upper unsigned __int32)
				l+=ADD_REG_IMM((unsigned char *)(codeptr+l),PC_PTR,Offset2);		// dst+(offset%8)
				l+=LOAD_REG_IMM((unsigned char *)(codeptr+l),NATIVE_2,Part1);
				l+=REP_MOVE_STORE_BYTE((unsigned char *)(codeptr+l));
				l+=POP_REGISTER((unsigned char *)(codeptr+l),MEM_PTR);
				l+=LOAD_REG_IMM((unsigned char *)(codeptr+l),NATIVE_2,Part2);
				l+=REP_MOVE_STORE_BYTE((unsigned char *)(codeptr+l));
				l+=POP_ALL((unsigned char *)(codeptr+l));
			}
			else
			{
				unsigned __int16 l=0;
				l+=PUSH_ALL((unsigned char *)(codeptr+l));
				l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
				l+=CLD((unsigned char *)(codeptr+l));
				l+=MAKE_PHYS_ADDR((unsigned char *)(codeptr+l),Offset,Imm);
				unsigned __int32 tmp=op0*8;
				l+=NATIVE_REG_TO_REG((unsigned char *)(codeptr+l),PC_PTR,REG_PTR);
				l+=ADD_REG_IMM((unsigned char *)(codeptr+l),PC_PTR,tmp);
				l+=ADD_REG_IMM((unsigned char *)(codeptr+l),PC_PTR,Offset2);		// dst+(offset%8)
				l+=LOAD_REG_IMM((unsigned char *)(codeptr+l),NATIVE_2,Part1);
				l+=REP_MOVE_STORE_BYTE((unsigned char *)(codeptr+l));
				l+=POP_ALL((unsigned char *)(codeptr+l));
			}
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
		else
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,Imm);
			l+=PUSH_REGISTER((unsigned char *)(codeptr+l),MEM_PTR);		//where
			l+=LOAD_REG_DWORD((unsigned char *)(codeptr+l),NATIVE_0,(unsigned char *)dynaHelperLDL);
			l+=CALL_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=POP_REGISTER((unsigned char *)(codeptr+l),NATIVE_1);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_0,op0*8);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_3,op0*8+4);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
	}
	_asm
	{
		mov esp,StackPointer	// nasty - get the original stack pointer; ie; before we called the BuilderFunction
								// this 'cleans' the stack.  We put five 32 bit values on the stack and the C++ compiler
								// put some stuff on it as well... we have no way of knowing how much, which is why we 
								// sent the original stackpointer in as a variable.
		popad					// all the original register, cause who knows what C++ did to them.... ;)
		add eax,0x50			// eax contains the starting address... return back to 0x50 past that
		push eax				// new return address
		ret						// hehe... bye!
	}
}


void dynaRuntimeBuilderLWL(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer)
{
	unsigned __int16 l=0;
	bool FlipFlopAddress=false;	
	unsigned __int32 address=m_Reg->r[op1*WIDTH]+Imm;
	unsigned __int32 LastAddress=*(unsigned __int32  *)((unsigned char *)codeptr+0x4c);
	unsigned __int32 NumAccess=*(unsigned __int32  *)((unsigned char *)codeptr+0x48);
	__int32 value;
	unsigned __int32 Mask;
	unsigned __int32 Offset;
	bool IsDirect;

	dynaHelperLWL(address,op0);

	*(unsigned __int32 *)((unsigned char *)codeptr+0x4c)=address;
	NumAccess--;
	*(unsigned __int32 *)((unsigned char *)codeptr+0x48)=NumAccess;
	dynaTranslateReadAddress(address,Imm,&Offset,&Mask,&IsDirect);
	if(((LastAddress&0xff000000)!=(address&0xff000000))&&(LastAddress!=0))
	{
		IsDirect=false;
		FlipFlopAddress=true;
	}
	if((NumAccess==0)||FlipFlopAddress)
	{
		if(IsDirect)//&&(Offset!=(unsigned __int32)dynaRamPtr))
		{
			int Count;
			unsigned __int32 Offset2;
			Offset2=Imm&7;
			Count=8-(int)Offset2;
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=PUSH_ALL((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=CLD((unsigned char *)(codeptr+l));
			l+=MAKE_PHYS_ADDR((unsigned char *)(codeptr+l),Offset,Imm);
			unsigned __int32 tmp=op0*8;
			l+=NATIVE_REG_TO_REG((unsigned char *)(codeptr+l),PC_PTR,REG_PTR);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),PC_PTR,tmp);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),PC_PTR,Offset2);		// dst+(offset%8)
			l+=LOAD_REG_IMM((unsigned char *)(codeptr+l),NATIVE_2,Count);
			l+=REP_MOVE_STORE_BYTE((unsigned char *)(codeptr+l));
			l+=POP_ALL((unsigned char *)(codeptr+l));
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
		else
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,Imm);
			l+=PUSH_REGISTER((unsigned char *)(codeptr+l),MEM_PTR);		//where
			l+=LOAD_REG_DWORD((unsigned char *)(codeptr+l),NATIVE_0,(unsigned char *)dynaHelperLWL);
			l+=CALL_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=POP_REGISTER((unsigned char *)(codeptr+l),NATIVE_1);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_0,op0*8);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
	}
	_asm
	{
		mov esp,StackPointer	// nasty - get the original stack pointer; ie; before we called the BuilderFunction
								// this 'cleans' the stack.  We put five 32 bit values on the stack and the C++ compiler
								// put some stuff on it as well... we have no way of knowing how much, which is why we 
								// sent the original stackpointer in as a variable.
		popad					// all the original register, cause who knows what C++ did to them.... ;)
		add eax,0x50			// eax contains the starting address... return back to 0x50 past that
		push eax				// new return address
		ret						// hehe... bye!
	}
}


void dynaRuntimeBuilderLDR(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer)
{
	unsigned __int16 l=0;
	bool FlipFlopAddress=false;	
	unsigned __int32 address=m_Reg->r[op1*WIDTH]+Imm;
	unsigned __int32 LastAddress=*(unsigned __int32  *)((unsigned char *)codeptr+0x4c);
	unsigned __int32 NumAccess=*(unsigned __int32  *)((unsigned char *)codeptr+0x48);
	__int32 value;
	unsigned __int32 Mask;
	unsigned __int32 Offset;
	bool IsDirect;

	dynaHelperLDR(address,op0);

	*(unsigned __int32 *)((unsigned char *)codeptr+0x4c)=address;
	NumAccess--;
	*(unsigned __int32 *)((unsigned char *)codeptr+0x48)=NumAccess;
	dynaTranslateReadAddress(address,Imm,&Offset,&Mask,&IsDirect);
	if(((LastAddress&0xff000000)!=(address&0xff000000))&&(LastAddress!=0))
	{
		IsDirect=false;
		FlipFlopAddress=true;
	}
	if((NumAccess==0)||FlipFlopAddress)
	{
		if(IsDirect)//&&(Offset!=(unsigned __int32)dynaRamPtr))
		{
			int Part1,Part2,Count;
			unsigned __int32 Offset2;
			Offset2=Imm&7;
			Count=8-(int)Offset2;
			Part1=Count-4;
			if(Count>=4) Part2=4;
			else Part2=Count;
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			if(Part1>0)
			{
				unsigned __int16 l=0;
				l+=PUSH_ALL((unsigned char *)(codeptr+l));
				l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
				l+=STD((unsigned char *)(codeptr+l));
				l+=MAKE_PHYS_ADDR((unsigned char *)(codeptr+l),Offset,Imm);
				unsigned __int32 tmp=op0*8;
				l+=NATIVE_REG_TO_REG((unsigned char *)(codeptr+l),PC_PTR,REG_PTR);
				l+=ADD_REG_IMM((unsigned char *)(codeptr+l),PC_PTR,tmp);
				l+=PUSH_REGISTER((unsigned char *)(codeptr+l),MEM_PTR);
				l+=ADD_REG_IMM((unsigned char *)(codeptr+l),PC_PTR,Offset2);		// dst+(offset%8)
				l+=LOAD_REG_IMM((unsigned char *)(codeptr+l),NATIVE_2,Part1);
				l+=REP_MOVE_STORE_BYTE((unsigned char *)(codeptr+l));
				l+=POP_REGISTER((unsigned char *)(codeptr+l),MEM_PTR);
				l+=LOAD_REG_IMM((unsigned char *)(codeptr+l),NATIVE_2,Part2);
				l+=REP_MOVE_STORE_BYTE((unsigned char *)(codeptr+l));
				l+=POP_ALL((unsigned char *)(codeptr+l));
				l+=CLD((unsigned char *)(codeptr+l));
			}
			else
			{
				unsigned __int16 l=0;
				l+=PUSH_ALL((unsigned char *)(codeptr+l));
				l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
				l+=STD((unsigned char *)(codeptr+l));
				l+=MAKE_PHYS_ADDR((unsigned char *)(codeptr+l),Offset,Imm);
				unsigned __int32 tmp=op0*8;
				l+=NATIVE_REG_TO_REG((unsigned char *)(codeptr+l),PC_PTR,REG_PTR);
				l+=ADD_REG_IMM((unsigned char *)(codeptr+l),PC_PTR,tmp);
				l+=ADD_REG_IMM((unsigned char *)(codeptr+l),PC_PTR,Offset2);		// dst+(offset%8)
				l+=LOAD_REG_IMM((unsigned char *)(codeptr+l),NATIVE_2,Part1);
				l+=REP_MOVE_STORE_BYTE((unsigned char *)(codeptr+l));
				l+=POP_ALL((unsigned char *)(codeptr+l));
				l+=CLD((unsigned char *)(codeptr+l));
			}
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
		else
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,Imm);
			l+=PUSH_REGISTER((unsigned char *)(codeptr+l),MEM_PTR);		//where
			l+=LOAD_REG_DWORD((unsigned char *)(codeptr+l),NATIVE_0,(unsigned char *)dynaHelperLDR);
			l+=CALL_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=POP_REGISTER((unsigned char *)(codeptr+l),NATIVE_1);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_0,op0*8);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_3,op0*8+4);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
	}
	_asm
	{
		mov esp,StackPointer	// nasty - get the original stack pointer; ie; before we called the BuilderFunction
								// this 'cleans' the stack.  We put five 32 bit values on the stack and the C++ compiler
								// put some stuff on it as well... we have no way of knowing how much, which is why we 
								// sent the original stackpointer in as a variable.
		popad					// all the original register, cause who knows what C++ did to them.... ;)
		add eax,0x50			// eax contains the starting address... return back to 0x50 past that
		push eax				// new return address
		ret						// hehe... bye!
	}
}

void dynaRuntimeBuilderLWR(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer)
{
	unsigned __int16 l=0;
	bool FlipFlopAddress=false;	
	unsigned __int32 address=m_Reg->r[op1*WIDTH]+Imm;
	unsigned __int32 LastAddress=*(unsigned __int32  *)((unsigned char *)codeptr+0x4c);
	unsigned __int32 NumAccess=*(unsigned __int32  *)((unsigned char *)codeptr+0x48);
	__int32 value;
	unsigned __int32 Mask;
	unsigned __int32 Offset;
	bool IsDirect;

	dynaHelperLWR(address,op0);

	*(unsigned __int32 *)((unsigned char *)codeptr+0x4c)=address;
	NumAccess--;
	*(unsigned __int32 *)((unsigned char *)codeptr+0x48)=NumAccess;
	dynaTranslateReadAddress(address,Imm,&Offset,&Mask,&IsDirect);
	if(((LastAddress&0xff000000)!=(address&0xff000000))&&(LastAddress!=0))
	{
		IsDirect=false;
		FlipFlopAddress=true;
	}
	if((NumAccess==0)||FlipFlopAddress)
	{
		if(IsDirect)//&&(Offset!=(unsigned __int32)dynaRamPtr))
		{
			int Count;
			unsigned __int32 Offset2;
			Offset2=Imm&3;
			Count=Offset2+1;
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=PUSH_ALL((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=STD((unsigned char *)(codeptr+l));
			l+=MAKE_PHYS_ADDR((unsigned char *)(codeptr+l),Offset,Imm);
			unsigned __int32 tmp=op0*8;
			l+=NATIVE_REG_TO_REG((unsigned char *)(codeptr+l),PC_PTR,REG_PTR);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),PC_PTR,tmp);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),PC_PTR,Offset2);		// dst+(offset%8)
			l+=LOAD_REG_IMM((unsigned char *)(codeptr+l),NATIVE_2,Count);
			l+=REP_MOVE_STORE_BYTE((unsigned char *)(codeptr+l));
			l+=POP_ALL((unsigned char *)(codeptr+l));
			l+=CLD((unsigned char *)(codeptr+l));
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
		else
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,Imm);
			l+=PUSH_REGISTER((unsigned char *)(codeptr+l),MEM_PTR);		//where
			l+=LOAD_REG_DWORD((unsigned char *)(codeptr+l),NATIVE_0,(unsigned char *)dynaHelperLWR);
			l+=CALL_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=POP_REGISTER((unsigned char *)(codeptr+l),NATIVE_1);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_0,op0*8);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
	}
	_asm
	{
		mov esp,StackPointer	// nasty - get the original stack pointer; ie; before we called the BuilderFunction
								// this 'cleans' the stack.  We put five 32 bit values on the stack and the C++ compiler
								// put some stuff on it as well... we have no way of knowing how much, which is why we 
								// sent the original stackpointer in as a variable.
		popad					// all the original register, cause who knows what C++ did to them.... ;)
		add eax,0x50			// eax contains the starting address... return back to 0x50 past that
		push eax				// new return address
		ret						// hehe... bye!
	}
}


void dynaRuntimeBuilderLDC2(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer)
{
	unsigned __int16 l=0;
	bool FlipFlopAddress=false;	
	unsigned __int32 address=m_Reg->r[op1*WIDTH]+Imm;
	unsigned __int32 LastAddress=*(unsigned __int32  *)((unsigned char *)codeptr+0x4c);
	unsigned __int32 NumAccess=*(unsigned __int32  *)((unsigned char *)codeptr+0x48);
	__int64 value;
	unsigned __int32 Mask;
	unsigned __int32 Offset;
	bool IsDirect;
	value=(__int64)helperLoadDWord(address);
	m_Reg->CPR[2][op0*WIDTH]=value;

	*(unsigned __int32 *)((unsigned char *)codeptr+0x4c)=address;
	NumAccess--;
	*(unsigned __int32 *)((unsigned char *)codeptr+0x48)=NumAccess;
	dynaTranslateReadAddress(address,Imm,&Offset,&Mask,&IsDirect);
	if(((LastAddress&0xff000000)!=(address&0xff000000))&&(LastAddress!=0))
	{
		IsDirect=false;
		FlipFlopAddress=true;
	}
	if((NumAccess==0)||FlipFlopAddress)
	{
		if(IsDirect)//&&(Offset!=(unsigned __int32)dynaRamPtr))
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=MAKE_PHYS_ADDR((unsigned char *)(codeptr+l),Offset,Imm);
			l+=LOAD_MEM_REG((unsigned char *)(codeptr+l),MEM_PTR,NATIVE_0);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_0,CPR2+op0*8+4);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,4);
			l+=LOAD_MEM_REG((unsigned char *)(codeptr+l),MEM_PTR,NATIVE_0);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_0,CPR2+op0*8);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
		else
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,Imm);
			l+=PUSH_REGISTER((unsigned char *)(codeptr+l),MEM_PTR);		//where
			l+=LOAD_REG_DWORD((unsigned char *)(codeptr+l),NATIVE_0,(unsigned char *)&helperLoadDWordAddress);
			l+=CALL_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=POP_REGISTER((unsigned char *)(codeptr+l),NATIVE_1);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_0,CPR2+op0*8);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_3,CPR2+op0*8+4);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
	}
	_asm
	{
		mov esp,StackPointer	// nasty - get the original stack pointer; ie; before we called the BuilderFunction
								// this 'cleans' the stack.  We put five 32 bit values on the stack and the C++ compiler
								// put some stuff on it as well... we have no way of knowing how much, which is why we 
								// sent the original stackpointer in as a variable.
		popad					// all the original register, cause who knows what C++ did to them.... ;)
		add eax,0x50			// eax contains the starting address... return back to 0x50 past that
		push eax				// new return address
		ret						// hehe... bye!
	}
}

void dynaRuntimeBuilderLWC2(unsigned __int32 codeptr,unsigned __int32 op0,unsigned __int32 op1,unsigned __int32 Imm,unsigned __int32 StackPointer)
{
	unsigned __int16 l=0;
	bool FlipFlopAddress=false;	
	unsigned __int32 address=m_Reg->r[op1*WIDTH]+Imm;
	unsigned __int32 LastAddress=*(unsigned __int32  *)((unsigned char *)codeptr+0x4c);
	unsigned __int32 NumAccess=*(unsigned __int32  *)((unsigned char *)codeptr+0x48);
	__int32 value;
	unsigned __int32 Mask;
	unsigned __int32 Offset;
	bool IsDirect;
	value=(__int32)helperLoadWord(address);
	m_Reg->CPR[2][op0*WIDTH]=(__int64)value;

	*(unsigned __int32 *)((unsigned char *)codeptr+0x4c)=address;
	NumAccess--;
	*(unsigned __int32 *)((unsigned char *)codeptr+0x48)=NumAccess;
	dynaTranslateReadAddress(address,Imm,&Offset,&Mask,&IsDirect);
	if(((LastAddress&0xff000000)!=(address&0xff000000))&&(LastAddress!=0))
	{
		IsDirect=false;
		FlipFlopAddress=true;
	}
	if((NumAccess==0)||FlipFlopAddress)
	{
		if(IsDirect)//&&(Offset!=(unsigned __int32)dynaRamPtr))
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=MAKE_PHYS_ADDR((unsigned char *)(codeptr+l),Offset,Imm);
			l+=LOAD_MEM_REG((unsigned char *)(codeptr+l),MEM_PTR,NATIVE_0);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_0,CPR2+op0*8);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
		else
		{
			l+=INC_PC_COUNTER((unsigned char *)(codeptr+l));
			l+=LOAD_REG((unsigned char *)(codeptr+l),op1,MEM_PTR);
			l+=ADD_REG_IMM((unsigned char *)(codeptr+l),MEM_PTR,Imm);
			l+=PUSH_REGISTER((unsigned char *)(codeptr+l),MEM_PTR);		//where
			l+=LOAD_REG_DWORD((unsigned char *)(codeptr+l),NATIVE_0,(unsigned char *)&helperLoadWordAddress);
			l+=CALL_REGISTER((unsigned char *)(codeptr+l),NATIVE_0);
			l+=POP_REGISTER((unsigned char *)(codeptr+l),NATIVE_1);
			l+=STORE_REG_TO_RBANK((unsigned char *)(codeptr+l),NATIVE_0,CPR2+op0*8);
			l+=JMP_SHORT((unsigned char *)(codeptr+l),0x50-(l+2));
		}
	}
	_asm
	{
		mov esp,StackPointer	// nasty - get the original stack pointer; ie; before we called the BuilderFunction
								// this 'cleans' the stack.  We put five 32 bit values on the stack and the C++ compiler
								// put some stuff on it as well... we have no way of knowing how much, which is why we 
								// sent the original stackpointer in as a variable.
		popad					// all the original register, cause who knows what C++ did to them.... ;)
		add eax,0x50			// eax contains the starting address... return back to 0x50 past that
		push eax				// new return address
		ret						// hehe... bye!
	}
}
